/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Klasse, die den Spielablauf eines Spiels, bei denen Textbausteine in die richtige Reihenfolge zu bringen sind, verwaltet, die Kommunikation mit der Lernplattform über SCORM-API übernimmt und die Anzeige der nötigen Elemente sicherstellt
*/
function ObjectcardTask(
    nodeParent,     // DOM-Knoten, als dessen einzige Kinder das Spiel angezeigt werden soll
    countNeeded     // Wie viele Richtige sind nötig?
) {
    ObjectcardTask.parent.call(this);
    nodeParent.innerHTML = '';
    this.countNeeded = countNeeded;
    this.msgPlaceGeneral = new ScoTools.MessagePlace(true);
    this.msgPlaceGeneral.appendAsChild(nodeParent);
    this.msgPlace = new ScoTools.MessagePlace();
    this.nodeParent = nodeParent;

    this.connect();
    this.robotconfig = ObjectcardTask.randomRobot();
    this._loadStatus();
    this.establish();
    AudioFeedbackTask.cache();
    this.ready = true;
}
ObjectcardTask.prototype = Object.create(AudioFeedbackTask.prototype);
ObjectcardTask.prototype.constructor = ObjectcardTask;
ObjectcardTask.parent = AudioFeedbackTask;

// Ist das Spiel gestartet?
ObjectcardTask.prototype.ready = false;

// Wie viele Richtige sind nötig, um die Aufgabe erfolgreich abzuschließen?
ObjectcardTask.prototype.countNeeded = 3;

// Wie viele Richtige wurden bereits erreicht?
ObjectcardTask.prototype.countDone = 0;

// Welche Roboterkonfiguration soll modelliert werden?
ObjectcardTask.prototype.robotconfig = null;

/**
* Aus Titelzeile der Objektkarte Objekt- und Klassenname extrahieren
* @return Object    { klasse: String, objekt: String, error: "String oder undefined" }
*/
ObjectcardTask.parseTitle = function(
    s       // Textinhalt der Titelzeile
) {
    var parts = s.split(':');
    switch (parts.length) {
        case 1:
            return { klasse: '', objekt: s.trim(), error: 'In Zeile 1 fehlt der Doppelpunkt.' };
        case 2:
            break;
        default:
            return { klasse: '', objekt: s.trim(), error: 'In Zeile 1 kann es nur einen Doppelpunkt geben.' };
    }
    var klasse = parts[1].trim();
    var objekt = parts[0].trim();
    if (!klasse.length)
        return { klasse: klasse, objekt: objekt, error: 'In Zeile 1 fehlt der Klassenname.' };
    var regVerboten = new RegExp('\\W');
    var arrVerboten = regVerboten.exec(objekt);
    if (arrVerboten && arrVerboten.length)
        return { klasse: klasse, objekt: objekt, error: 'Objektname in Zeile 1 enthält verbotenes Zeichen: ' + arrVerboten[0] };
    arrVerboten = regVerboten.exec(klasse);
    if (arrVerboten && arrVerboten.length)
        return { klasse: klasse, objekt: objekt, error: 'Klassenname in Zeile 1 enthält verbotenes Zeichen: ' + arrVerboten[0] };
    return { klasse: klasse, objekt: objekt };
}

/**
* Aus Objektkarte Attribute auslesen
* @return mixed    Gelesene Werte als Objekt oder String mit Fehlermeldung
*/
ObjectcardTask.parseAttributes = function(
    s,          // Attributaufzählung auf Objektkarte
    arrValid    // Liste mit erlaubten Attributnamen
) {
    var obj = {};
    var arrInfos = s.split('\n');
    var s, parts, attname, attvalue;
    for (var z = 0; z < arrInfos.length; ++z) {
        s = arrInfos[z].trim();
        if (!s.length)
            continue;
        parts = s.split('=');
        switch (parts.length) {
            case 1:
                return 'Gleichheitszeichen fehlt in Zeile ' + (z + 2) + '.';
            case 2:
                break;
            default:
                return 'In Zeile ' + (z + 2) + ' kann es nicht mehrere Gleichheitszeichen geben.';
        }
        attname = parts[0].trim();
        attvalue = parts[1].trim();
        if (arrValid.indexOf(attname) == -1)
            return 'Fehler in Zeile ' + (z + 2) + ': Ein Attribut namens „' + attname + '“ ist nicht vorgesehen.';
        if (!attvalue.length)
            return 'In Zeile ' + (z + 2) + ' fehlt beim Attribut „' + attname + '“ der Attributwert.';
        if (obj[attname] !== undefined)
            return 'Zum Attribut „' + attname + '“ darf nicht mehrfach ein Wert angegeben werden.';
        obj[attname] = attvalue;
    }
    return obj;
}

/**
* Ermittelt die gespeicherte Eingabe
* @return Object    Objekt der Form { title: "Inhalt der obersten Zeile der Objektkarte", atts: "Inhalt des unteren Bereichs der Objektkarte" }
*/
ObjectcardTask.prototype._getSaved = function() {
    if (this.isConnected()) {
        var data = this.read('suspendData');
        if (data) {
            data = data.split('\n');
            return { title: data.length ? data.shift() : '', atts: data.join('\n') };
        }
    }
    return { title: '', atts: '' };
}

/**
* Lädt den Zwischenstand vom LMS
* @return undefined
*/
ObjectcardTask.prototype._loadStatus = function() {
    if (this.isConnected()) {
        var data = this.read('lessonLocation');
        if (data) {
            data = data.split('|');
            // Anzahl geschaffter Aufgaben
            var n = +data.shift();
            if (Number.isFinite(n))
                this.countDone = Math.min(this.countNeeded, Math.max(0, n));
            // Farbe?
            n = data.length ? data.shift() : 'keine';
            if (createrobotimg.ARR_FARBEN.indexOf(n) != -1)
                this.robotconfig.farbe = n;
            // Achsenanzahl
            n = data.length ? (+data.shift()) : undefined;
            if (Number.isFinite(n) && !(n % 1))
                this.robotconfig.achsen = Math.max(0, Math.min(5, n));
            // Höhe
            n = data.length ? (+data.shift()) : undefined;
            if (Number.isFinite(n) && !(n % 1))
                this.robotconfig.hoehe = Math.max(1, Math.min(10, n));
        }
    }
}

/**
* Aktuellen Status speichern
* @return undefined
*/
ObjectcardTask.prototype.doSave = function() {
    if (this.getDone()) {
        this.reportTime();
        return;
    }
    if (!this.lmslink.ready)
        return;
    this.save({
        'score': Math.min(100, Math.max(0, this.countDone / this.countNeeded * 100)),
        'scoreMin': 0,
        'scoreMax': 100,
        'lessonStatus': (this.countDone >= this.countNeeded) ? 'completed' : 'incomplete',
        'lessonLocation': this.countDone + '|' + this.robotconfig.farbe + '|' + this.robotconfig.achsen + '|' + this.robotconfig.hoehe,
        'suspendData': this._okt.value + '\n' + this._oka.value
    });
}

/**
* Eine zufällige Roboterkonfiguration erstellen
* @return Object        Roboterkonfiguration
*/
ObjectcardTask.randomRobot = function() {
    return {
        farbe: createrobotimg.ARR_FARBEN[Math.floor(Math.random() * createrobotimg.ARR_FARBEN.length)],
        achsen: Math.floor(Math.random() * 4) + 2,
        hoehe: Math.floor(Math.random() * 10) + 1
    };
}

/**
* Ggf. ein Roboterbild erstellen und anzeigen
* @return undefined
*/
ObjectcardTask.showRobot = function(
    nodeFigure,     // Figure-Element, in dem evtl. Roboterbild angezeigt werden soll
    objAtts         // Optional: Roboterkonfiguration
) {
    if (nodeFigure.firstChild != nodeFigure.lastChild)
        nodeFigure.removeChild(nodeFigure.lastChild);
    if (objAtts)
        ScoTools.createHtml('img', nodeFigure, {
            'alt': 'Aussehen des Roboters',
            'src': createrobotimg(objAtts.farbe, objAtts.achsen, objAtts.hoehe)
        });
}

/**
* Bedienelemente für das Lernpaket aufbauen
* @return undefined
*/
ObjectcardTask.prototype.establish = function() {
    ObjectcardTask.parent.prototype.establish.call(this);
    ScoTools.createHtml('h2', this.nodeParent, null, 'Objektkarte');
    ScoTools.createHtml('p', this.nodeParent, null, 'Fülle die Objektkarte so aus, dass der als „Tatsächlicher Roboter“ gezeigte Roboter durch die Objektkarte vollständig beschrieben wird:');

    var imgplace = ScoTools.createHtml('div', ScoTools.createHtml('div', this.nodeParent, { 'class': 'imgplace' }));
    this._figureSoll = ScoTools.createHtml('figure', imgplace);
    ScoTools.createHtml('figcaption', this._figureSoll, null, 'Tatsächlicher Roboter:');
    this._figureIst = ScoTools.createHtml('figure', imgplace);
    ScoTools.createHtml('figcaption', this._figureIst, null, 'Laut Objektkarte:');

    this.msgPlace.appendAsChild(this.nodeParent);

    var self = this;
    var form = ScoTools.createHtml('form', ScoTools.createHtml('div', this.nodeParent, { 'style': 'text-align:center' }), { 'method': 'post', 'action': location.href, 'class': 'objektkarte' });
    form.addEventListener('submit', function(ev) { ev.preventDefault; ev.stopPropagation });
    form.addEventListener('input', function() { self.procChange(); });
    form.addEventListener('change', function() { self.doSave(); });
    var data = this._getSaved();
    this._okt = ScoTools.createHtml('input', form, { 'placeholder': 'Name: TR', 'value': data.title, 'autocomplete': 'off' });
    this._oka = ScoTools.createHtml('textarea', form, { 'placeholder': 'farbe = …\nachsen = …\nhoehe = …', 'autocomplete': 'off' }, data.atts);

    ObjectcardTask.showRobot(this._figureSoll, this.robotconfig);
    this.procChange();
}

/**
* Einen menschenlesbaren String mit noch nötiger Objektkartenanzahl erstellen
* @return String    Erstellter String
*/
ObjectcardTask.prototype.howMuchLeftString = function() {
    if (this.countDone >= this.countNeeded)
        return 'Natürlich kannst du weiterüben, obwohl du diese Aufgabe bereits erfolgreich abgeschlossen hast';
    var s = 'Fülle noch ';
    switch (this.countNeeded - this.countDone) {
        case 1:
            s += 'eine Objektkarte';
            break;
        case 2:
            s += 'zwei Objektkarten';
            break;
        case 3:
            s += 'drei Objektkarten';
            break;
        case 4:
            s += 'vier Objektkarten';
            break;
        case 5:
            s += 'fünf Objektkarten';
            break;
        case 6:
            s += 'sechs Objektkarten';
            break;
        case 7:
            s += 'sieben Objektkarten';
            break;
        case 8:
            s += 'acht Objektkarten';
            break;
        case 9:
            s += 'neun Objektkarten';
            break;
        case 10:
            s += 'zehn Objektkarten';
            break;
        case 11:
            s += 'elf Objektkarten';
            break;
        case 12:
            s += 'zwölf Objektkarten';
            break;
        default:
            s += (this.countNeeded - this.countDone) + ' Objektkarten';
    }
    return s + ' korrekt aus';
}

/**
* Festsetzen, ob die Werte in der Objektkarte gültig sind und ggf. Fehlermeldung anzeigen
* @return undefined
*/
ObjectcardTask.prototype._reportError = function(
    strMsg,         // Anzuzeigende Meldung
    boolTitle       // Optional: Ist der Titel das Problem?
) {
    if (strMsg)
        this.showMsg(strMsg, ScoTools.MSGTYPE_WARN);
    if (this._okt.setCustomValidity)
        this._okt.setCustomValidity((!strMsg || !boolTitle) ? '' : strMsg);
    if (this._oka.setCustomValidity)
        this._oka.setCustomValidity((!strMsg || boolTitle) ? '' : strMsg);
}

/**
* Darauf reagieren, wenn Objektkarte verändert wird
* @return undefined
*/
ObjectcardTask.prototype.procChange = function() {
    var title = this._okt.value;
    var atts = this._oka.value;
    // Speichern!
    //this.doSave();

    ObjectcardTask.showRobot(this._figureIst);
    // Objektkarte leer?
    if (!(title + atts).length) {
        this.showMsg(this.howMuchLeftString(), ScoTools.MSGTYPE_INFO);
        return;
    }
    // Titelzeile OK?
    var title = ObjectcardTask.parseTitle(title);
    if (title.error) {
        this._reportError(title.error, true);
        return;
    }
    if (title.klasse != 'TR') {
        this._reportError('Angegebene Klasse „' + title.klasse + '“ ist unbekannt', true);
        return;
    }
    // Attribute prinzipiell OK?
    atts = ObjectcardTask.parseAttributes(atts, ['farbe', 'achsen', 'hoehe']);
    if ((typeof atts == 'string') || (atts instanceof String)) {
        this._reportError(atts);
        return;
    }
    // Passen die Farbangaben?
    var ATTS_COLORS = ['farbe'];
    for (z = 0; z < ATTS_COLORS.length; ++z) {
        if (atts[ATTS_COLORS[z]] === undefined) {
            this._reportError('Das Attribut „' + ATTS_COLORS[z] + '“ fehlt');
            return;
        }
        if ((atts[ATTS_COLORS[z]] = createrobotimg.parseColor(atts[ATTS_COLORS[z]])) === undefined) {
            this._reportError('Wert für Attribut „' + ATTS_COLORS[z] + '“ ist ungültig. (Gültige Farben sind rot, grün, blau und grau.)');
            return;
        }
    }
    // Passen die nummerischen Werte?
    var ATTS_NUMBERIC = [ { key: 'achsen', min: 2, max: 5 }, { key: 'hoehe', min: 1, max: 20} ];
    for (var z = 0; z < ATTS_NUMBERIC.length; ++z) {
        if (atts[ATTS_NUMBERIC[z].key] === undefined) {
            this._reportError('Das Attribut „' + ATTS_NUMBERIC[z].key + '“ fehlt');
            return;
        }
        if (!(new RegExp('^[+\\-]?\\d+$').test(atts[ATTS_NUMBERIC[z].key]))) {
            this._reportError('Wert für Attribut „' + ATTS_NUMBERIC[z].key + '“ ist ungültig');
            return;
        }
        atts[ATTS_NUMBERIC[z].key] = +atts[ATTS_NUMBERIC[z].key];
        if (atts[ATTS_NUMBERIC[z].key] < ATTS_NUMBERIC[z].min) {
            this._reportError('Wert für Attribut „' + ATTS_NUMBERIC[z].key + '“ ist zu klein (Mindestwert ' + ATTS_NUMBERIC[z].min + ')');
            return;
        }
        if (atts[ATTS_NUMBERIC[z].key] > ATTS_NUMBERIC[z].max) {
            this._reportError('Wert für Attribut „' + ATTS_NUMBERIC[z].key + '“ ist zu groß (Maximalwert ' + ATTS_NUMBERIC[z].max + ')');
            return;
        }
    }
    // Alles als Fehlerfrei markieren
    this._reportError();
    ObjectcardTask.showRobot(this._figureIst, atts);

    // Passt die Objektkarte komplett?
    if (ObjectcardTask.isEqual(atts, this.robotconfig))
        this.reportCorrect();
    else
        this.showMsg('Die Objektkarte passt nicht zum tatsächlichen Roboter…');
}

/**
* Prüft, ob zwei Roboterkonfigurationen übereinstimmen
* @return Boolean   Übereinstimmung?
*/
ObjectcardTask.isEqual = function(
    c1,     // Roboterkonfiguration 1
    c2      // Roboterkonfiguration 2
) {
    return ((createrobotimg.parseColor(c1.farbe) == createrobotimg.parseColor(c2.farbe)) && (c1.achsen == c2.achsen) && (c1.hoehe == c2.hoehe));
}

/**
* Eine Meldung über das Ergebnis der Antwortüberprüfung anzeigen
* @return undefined
*/
ObjectcardTask.prototype.reportCorrect = function() {
    ++this.countDone;
    // Erfolgsmeldung
    this.showMsg((this.countDone < this.countNeeded) ? ('Richtig: ' + this.howMuchLeftString()) : 'Du hast die Objektkarte fehlerfrei ausgefüllt', true);
    this.showFinalMessage();
    // Neue Aufgabe
    this.robotconfig = ObjectcardTask.randomRobot();
    this._okt.value = '';
    this._oka.value = '';
    this._reportError();
    // Speichern
    this.doSave();
    // Neuen Roboter anzeigen
    ObjectcardTask.showRobot(this._figureSoll, this.robotconfig);
    ObjectcardTask.showRobot(this._figureIst);
    // Erfolgssound
    if (this.ready)
        this.playAudio(true);
    // Titelzeile fokusieren
    this._okt.focus();
}
